from typing import Any, Literal, NotRequired, TypedDict

from .accessibility import Accessibility
from .image import Image, Thumbnails
from .runs import Runs
from .simpletext import SimpleText
from .tracking import ClickTrackingParams, TrackingParams
from .urlendpoint import ContextMenuEndpoint


class Icon(TypedDict):
    iconType: Literal["OWNER", "MODERATOR"]


class LiveChatAuthorBadgeRenderer(TypedDict):
    customThumbnail: Thumbnails
    tooltip: str
    accessibility: Accessibility
    icon: NotRequired[Icon]


class AuthorBadge(TypedDict):
    liveChatAuthorBadgeRenderer: LiveChatAuthorBadgeRenderer


class ClientResource(TypedDict):
    imageName: str


class Source(TypedDict):
    clientResource: ClientResource


class Sources(TypedDict):
    sources: list[Source]


class ImageTint(TypedDict):
    color: int


class BorderImageProcessor(TypedDict):
    imageTint: ImageTint


class Processor(TypedDict):
    bprderImageProcessor: BorderImageProcessor


class UnheartedIcon(TypedDict):
    sources: list[Source]
    processor: Processor


class CreatorHeartViewModel(TypedDict):
    creatorThumbnail: Thumbnails
    heartedIcon: Sources
    unheartedIcon: UnheartedIcon
    heartedHoverText: str
    heartedAccessibilityLabel: str
    unheartedAccessibilityLabel: str
    engagementStateKey: str


class CreatorHeartButton(TypedDict):
    creatorHeartViewModel: CreatorHeartViewModel


class LiveChatRenderer(TypedDict):
    id: str
    timestampUsec: str
    authorExternalChannelId: str
    message: Runs


class AuthorInfo(TypedDict):
    authorName: SimpleText
    authorPhoto: Thumbnails
    authorBadges: NotRequired[list[AuthorBadge]]


class LiveChatTextMessageRenderer(LiveChatRenderer, AuthorInfo):
    contextMenuEndpoint: ContextMenuEndpoint
    contextMenuAccessibility: Accessibility


class LiveChatPaidMessageRenderer(TrackingParams, LiveChatRenderer, AuthorInfo):
    purchaseAmountText: SimpleText
    headerBackgroundColor: int
    headerTextColor: int
    bodyBackgroundColor: int
    bodyTextColor: int
    authorNameTextColor: int
    contextMenuEndpoint: ContextMenuEndpoint
    timestampColor: int
    contextMenuAccessibility: Accessibility
    textInputBackgroundColor: int
    creatorHeartButton: CreatorHeartButton
    isV2Style: bool


class LiveChatPaidStickerRenderer(TrackingParams, LiveChatRenderer, AuthorInfo):
    sticker: Image
    purchaseAmountText: SimpleText
    contextMenuEndpoint: ContextMenuEndpoint
    contextMenuAccessibility: Accessibility


class LiveChatMembershipItemRenderer(LiveChatRenderer, AuthorInfo):
    headerSubtext: Runs


class LiveChatSponsorshipsHeaderRenderer(AuthorInfo):
    """
    "liveChatSponsorshipsHeaderRenderer": {
        "authorName": {
            "simpleText": "\u267e\ufe0f\u91ce\u3046\u3055\u304e"
        },
        "authorPhoto": {
            "thumbnails": [
                {
                    "url": "https://yt4.ggpht.com/-Rp0B3c4BDQcB71RKqitQdCu2L7h3EqNNqdoqPWvRC-TguuzDUztmy1hTSpqQeEC5RLqsgn3fyw=s32-c-k-c0x00ffffff-no-rj",
                    "width": 32,
                    "height": 32
                },
                {
                    "url": "https://yt4.ggpht.com/-Rp0B3c4BDQcB71RKqitQdCu2L7h3EqNqdoqPWvRC-TguuzDUztmy1hTSpqQeEC5RLqsgn3fyw=s64-c-k-c0x00ffffff-no-rj",
                    "width": 64,
                    "height": 64
                }
            ]
        },
        "primaryText": {
            "runs": [
                {
                    "text": "Gifted ",
                    "bold": true
                },
                {
                    "text": "5",
                    "bold": true
                },
                {
                    "text": " ",
                    "bold": true
                },
                {
                    "text": "Pekora Ch. \u514e\u7530\u307a\u3053\u3089",
                    "bold": true
                },
                {
                    "text": " memberships",
                    "bold": true
                }
            ]
        },
        "authorBadges": [
            {
                "liveChatAuthorBadgeRenderer": {
                    "customThumbnail": {
                        "thumbnails": [
                            {
                                "url": "https://yt3.ggpht.com/ikjRH2-DarXi4D9rQptqzbl34YrHSkAs7Uyq41itvqRiYYcpq2zNYC2scrZ9gbXQEhBuFfOZuw=s16-c-k",
                                "width": 16,
                                "height": 16
                            },
                            {
                                "url": "https://yt3.ggpht.com/ikjRH2-DarXi4D9rQptqzbl34YrHSkAs7Uyq41itvqRiYYcpq2zNYC2scrZ9gbXQEhBuFfOZuw=s32-c-k",
                                "width": 32,
                                "height": 32
                            }
                        ]
                    },
                    "tooltip": "Member (2 months)",
                    "accessibility": {
                        "accessibilityData": {
                            "label": "Member (2 months)"
                        }
                    }
                }
            }
        ],
        "contextMenuEndpoint": {
            "clickTrackingParams": "CAUQ3MMKIhMIgbSe1t6qhAMVVkP1BR04yA_O",
            "commandMetadata": {
                "webCommandMetadata": {
                    "ignoreNavigation": true
                }
            },
            "liveChatItemContextMenuEndpoint": {
                "params": "Q2g0S0hBb2FRMDlZUlRVNFJHVnhiMUZFUm1GNlJYZG5VV1JVWjAxTFRYY2FLU29uQ2hoVlF6RkVRMlZrVW1kSFNFSmtiVGd4UlRGc2JFeG9UMUVTQ3pOa2FIRlFWRXhzTkRoQklBSW9CRElhQ2hoVlEwcGpVbnA1Umw4MVNYRkxkM1ZsZW5WNmFXUTFaVkU0QWtnQVVDUSUzRA=="
            }
        },
        "contextMenuAccessibility": {
            "accessibilityData": {
                "label": "Chat actions"
            }
        },
        "image": {
            "thumbnails": [
                {
                    "url": "https://www.gstatic.com/youtube/img/sponsorships/sponsorships_gift_purchase_announcement_artwork.png"
                }
            ]
        }
    }
    """

    primaryText: Runs
    contextMenuEndpoint: ContextMenuEndpoint
    contextMenuAccessibility: Accessibility
    image: Image


class LiveChatSponsorshipsGiftPurchaseAnnouncementRendererHeader(TypedDict):
    liveChatSponsorshipsHeaderRenderer: LiveChatSponsorshipsHeaderRenderer


class LiveChatSponsorshipsGiftPurchaseAnnouncementRenderer(LiveChatRenderer):
    """
    {
        "liveChatSponsorshipsGiftPurchaseAnnouncementRenderer": {
            "id": "ChwKGkNPWEU1OERlcW9RREZhekV3Z1FkVGdNS013",
            "timestampUsec": "1707910568302677",
            "authorExternalChannelId": "UCJcRzyF_5IqKwuezuzid5eQ",
            "header": LiveChatSponsorshipsGiftPurchaseAnnouncementRendererHeader
        }
    }
    """

    header: LiveChatSponsorshipsGiftPurchaseAnnouncementRendererHeader


class AddChatItemActionItem(TypedDict):
    liveChatTextMessageRenderer: NotRequired[LiveChatTextMessageRenderer]
    liveChatPaidMessageRenderer: NotRequired[LiveChatPaidMessageRenderer]
    liveChatPaidStickerRenderer: NotRequired[LiveChatPaidStickerRenderer]
    liveChatMembershipItemRenderer: NotRequired[LiveChatMembershipItemRenderer]
    liveChatSponsorshipsGiftRedemptionAnnouncementRenderer: NotRequired[
        LiveChatTextMessageRenderer
    ]
    liveChatSponsorshipsGiftPurchaseAnnouncementRenderer: NotRequired[
        LiveChatSponsorshipsGiftPurchaseAnnouncementRenderer
    ]


class AddChatItemAction(TypedDict):
    item: AddChatItemActionItem


class MarkChatItemAsDeletedAction(TypedDict):
    deletedStateMessage: Runs
    targetItemId: str


class AddLiveChatTickerItemActionItem(TypedDict):
    liveChatTickerPaidMessageItemRenderer: dict


class AddLiveChatTickerItemAction(TypedDict):
    item: AddLiveChatTickerItemActionItem


class ChoicesItem(TypedDict):
    text: Runs
    selected: bool
    voteRatio: float
    votePercentage: SimpleText
    signinEndpoint: Any


class PollHeaderRenderer(TypedDict):
    pollQuestion: Runs
    thumbnail: Thumbnails
    metadataText: Runs
    liveChatPollType: Literal["LIVE_CHAT_POLL_TYPE_CREATOR"]
    contextMenuButton: Any


class PollRendererHeader(TypedDict):
    pollHeaderRenderer: PollHeaderRenderer


class PollRenderer(TypedDict):
    choices: list[ChoicesItem]
    liveChatPollId: str
    header: PollRendererHeader


class PollToUpdate(TypedDict):
    pollRenderer: PollRenderer


class UpdateLiveChatPollAction(TypedDict):
    pollToUpdate: PollToUpdate


class ChatActionsItem(ClickTrackingParams):
    addChatItemAction: NotRequired[AddChatItemAction]
    removeChatItemByAuthorAction: NotRequired[dict]
    markChatItemAsDeletedAction: NotRequired[MarkChatItemAsDeletedAction]
    addLiveChatTickerItemAction: NotRequired[AddLiveChatTickerItemAction]
    updateLiveChatPollAction: NotRequired[UpdateLiveChatPollAction]


type ChatActions = list[ChatActionsItem]



from .frameworkupdates import FrameworkUpdates
from .responsecontext import ResponseContext
from .tracking import TrackingParams


class YoutubeResponse(TrackingParams):
    responseContext: ResponseContext
    frameworkUpdates: FrameworkUpdates



from __future__ import annotations

import json
import re
from datetime import timedelta
from typing import Any

import aiohttp
import bs4
from omu import Omu
from omu_chatprovider.errors import ProviderError
from omu_chatprovider.helper import assert_none
from omu_chatprovider.throttle import Throttle

from . import types
from .const import (
    BASE_HEADERS,
    BASE_PAYLOAD,
    YOUTUBE_REGEX,
    YOUTUBE_URL,
)


class YoutubePage:
    def __init__(
        self,
        soup: bs4.BeautifulSoup,
    ):
        self.soup = soup

    @classmethod
    async def from_response(cls, response: aiohttp.ClientResponse) -> YoutubePage:
        response_text = await response.text()
        soup = bs4.BeautifulSoup(response_text, "html.parser")
        return cls(soup)

    def get_ytcfg(self) -> types.ytcfg:
        ytcfg_data = self.extract_script("ytcfg.set")
        return assert_none(
            ytcfg_data,
            "Could not find ytcfg data",
        )

    def get_ytinitialdata(self) -> types.ytinitialdata:
        initial_data = self.extract_script('window["ytInitialData"]')
        return assert_none(
            initial_data,
            "Could not find initial data",
        )

    def get_ytinitialplayerresponse(self) -> types.ytInitialPlayerResponse:
        initial_player_response = self.extract_script("var ytInitialPlayerResponse")
        return assert_none(
            initial_player_response,
            "Could not find initial player response",
        )

    @property
    def INNERTUBE_API_KEY(self) -> str:
        return self.get_ytcfg()["INNERTUBE_API_KEY"]

    def extract_script(self, prefix: str) -> Any | None:
        for script in self.soup.select("script"):
            script_text = script.text.strip()
            if script_text.startswith(prefix):
                break
        else:
            return None
        if "{" not in script_text or "}" not in script_text:
            return None
        data_text = script_text[script_text.index("{") : script_text.rindex("}") + 1]
        data = json.loads(data_text)
        return data


class YoutubeAPI:
    def __init__(self, omu: Omu, session: aiohttp.ClientSession):
        self.omu = omu
        self.session = session
        self.throttle = Throttle(timedelta(seconds=1 / 3))

    async def get(
        self,
        url: str,
        params: dict[str, str] | None = None,
    ) -> YoutubePage:
        async with self.throttle:
            response = await self.session.get(
                url,
                params=params,
                headers=BASE_HEADERS,
            )
        return await YoutubePage.from_response(response)

    async def fetch_online_videos(self, url: str) -> list[str]:
        match = assert_none(
            re.search(YOUTUBE_REGEX, url),
            "Could not match url",
        )
        options = match.groupdict()

        video_id = options.get("video_id") or options.get("video_id_short")
        if video_id is None:
            channel_id = options.get(
                "channel_id"
            ) or await self.get_channel_id_by_vanity(
                options.get("channel_id_vanity")
                or options.get("channel_id_user")
                or options.get("channel_id_c")
            )
            if channel_id is None:
                raise ProviderError("Could not find channel id")
            video_id = await self.get_video_id_by_channel(channel_id)
            if video_id is None:
                return []
        if not await self.is_online(video_id):
            return []
        return [video_id]

    async def get_channel_id_by_vanity(self, vanity: str | None) -> str | None:
        if vanity is None:
            return None
        clean_vanity = re.sub(r"[^a-zA-Z0-9_-]", "", vanity)
        if not clean_vanity:
            return None
        async with self.throttle:
            response = await self.session.get(f"{YOUTUBE_URL}/@{clean_vanity}")
        soup = bs4.BeautifulSoup(await response.text(), "html.parser")
        meta_tag = soup.select_one('meta[itemprop="identifier"]')
        if meta_tag is None:
            return None
        return meta_tag.attrs.get("content")

    async def get_video_id_by_channel(self, channel_id: str) -> str | None:
        async with self.throttle:
            response = await self.session.get(
                f"{YOUTUBE_URL}/embed/live_stream?channel={channel_id}",
                headers=BASE_HEADERS,
            )
        soup = bs4.BeautifulSoup(await response.text(), "html.parser")
        canonical_link = soup.select_one('link[rel="canonical"]')
        if canonical_link is None:
            return await self.get_video_id_by_channel_feeds(channel_id)
        href = canonical_link.attrs.get("href")
        if href is None:
            return None
        match = re.search(YOUTUBE_REGEX, href)
        if match is None:
            return None
        options = match.groupdict()
        return options.get("video_id") or options.get("video_id_short")

    async def get_video_id_by_channel_feeds(self, channel_id: str) -> str | None:
        async with self.throttle:
            response = await self.session.get(
                f"{YOUTUBE_URL}/feeds/videos.xml?channel_id={channel_id}",
                headers=BASE_HEADERS,
            )
        soup = bs4.BeautifulSoup(await response.text(), "xml")
        link = soup.select_one("entry link")
        if link is None:
            return None
        href = link.attrs.get("href")
        if href is None:
            return None
        match = re.search(YOUTUBE_REGEX, href)
        if match is None:
            return None
        options = match.groupdict()
        return options.get("video_id") or options.get("video_id_short")

    async def get_live_chat(
        self,
        /,
        video_id: str,
        key: str,
        continuation: str | None = None,
    ) -> types.live_chat:
        url = f"{YOUTUBE_URL}/youtubei/v1/live_chat/get_live_chat"
        params = {
            "v": video_id,
            "key": key,
        }
        payload: dict[str, Any] = {
            **BASE_PAYLOAD,
        }
        if continuation:
            payload["continuation"] = continuation
        async with self.throttle:
            response = await self.session.post(
                url,
                params=params,
                headers=BASE_HEADERS,
                json=payload,
            )
        return await response.json()

    async def updated_metadata(
        self,
        /,
        video_id: str,
        key: str,
        continuation: str | None = None,
    ) -> types.updated_metadata:
        url = f"{YOUTUBE_URL}/youtubei/v1/updated_metadata"
        params = {
            "key": key,
        }
        payload: dict[str, Any] = {
            **BASE_PAYLOAD,
        }
        if continuation is not None:
            payload["continuation"] = continuation
        elif video_id is not None:
            payload["videoId"] = video_id
        else:
            raise ValueError("video_id or continuation must be provided")
        async with self.throttle:
            response = await self.session.post(
                url,
                params=params,
                headers=BASE_HEADERS,
                json=payload,
            )
        return await response.json()

    async def is_online(self, video_id: str) -> bool:
        live_chat_params = {"v": video_id}
        async with self.throttle:
            live_chat_response = await self.session.get(
                f"{YOUTUBE_URL}/live_chat",
                params=live_chat_params,
                headers=BASE_HEADERS,
            )
        if live_chat_response.status // 100 != 2:
            return False
        response = await YoutubePage.from_response(live_chat_response)
        initial_data = response.get_ytinitialdata()
        continuation = (
            initial_data.get("contents", {})
            .get("liveChatRenderer", {})
            .get("continuations", [{}])[0]
            .get("invalidationContinuationData", {})
            .get("continuation")
        )
        if continuation is None:
            return False
        live_chat_response_data = await self.get_live_chat(
            video_id=video_id,
            key=response.get_ytcfg()["INNERTUBE_API_KEY"],
            continuation=continuation,
        )
        return "continuationContents" in live_chat_response_data
